@@ -1,6 +1,6 @@
 /*  plugins/setGT.c -- set gentoypes to given values
 
-    Copyright (C) 2015-2021 Genome Research Ltd.
+    Copyright (C) 2015-2024 Genome Research Ltd.
 
     Author: Petr Danecek <pd3@sanger.ac.uk>
 
@@ -27,6 +27,7 @@ DEALINGS IN THE SOFTWARE.  */
 #include <htslib/vcf.h>
 #include <htslib/vcfutils.h>
 #include <htslib/kfunc.h>
+#include <htslib/hts_os.h>
 #include <inttypes.h>
 #include <getopt.h>
 #include <ctype.h>
@@ -48,19 +49,19 @@ static int cmp_gt(double a, double b) { return a>b ? 1 : 0; }
 typedef struct
 {
     bcf_hdr_t *in_hdr, *out_hdr;
-    int32_t *gts, mgts, *iarr, miarr;
+    int32_t *gts, mgts, *iarr, miarr, *xarr, mxarr;
     int *arr, marr;
     uint64_t nchanged;
     int tgt_mask, new_mask, new_gt;
     filter_t *filter;
     char *filter_str;
     struct {
-        int m_allele, M_allele, *gt, *phased, ploidy;
+        int m_allele, M_allele, x_vaf_allele, *gt, *phased, ploidy;
         char *gt_str;
     } custom;
-    int filter_logic;
+    int filter_logic, rand_seed;
     uint8_t *smpl_pass;
-    double binom_val;
+    double binom_val, rand_frac;
     char *binom_tag;
     cmp_f binom_cmp;
 }
@@ -79,9 +80,12 @@ args_t *args = NULL;
 #define GT_BINOM    (1<<8)
 #define GT_MINOR    (1<<9)
 #define GT_CUSTOM   (1<<10)
+#define GT_X_VAF    (1<<11)
+#define GT_RAND     (1<<12)
 
 #define MINOR_ALLELE -1
 #define MAJOR_ALLELE -2
+#define X_VAF_ALLELE -3
 
 const char *about(void)
 {
@@ -90,31 +94,34 @@ const char *about(void)
 
 const char *usage(void)
 {
-    return 
+    return
         "About: Sets genotypes. The target genotypes can be specified as:\n"
-        "           ./.  .. completely missing (\".\" or \"./.\", depending on ploidy)\n"
-        "           ./x  .. partially missing (e.g., \"./0\" or \".|1\" but not \"./.\")\n"
-        "           .    .. partially or completely missing\n"
-        "           a    .. all genotypes\n"
-        "           b    .. heterozygous genotypes failing two-tailed binomial test (example below)\n"
-        "           q    .. select genotypes using -i/-e options\n"
+        "           ./.     .. completely missing (\".\" or \"./.\", depending on ploidy)\n"
+        "           ./x     .. partially missing (e.g., \"./0\" or \".|1\" but not \"./.\")\n"
+        "           .       .. partially or completely missing\n"
+        "           a       .. all genotypes\n"
+        "           b       .. heterozygous genotypes failing two-tailed binomial test (example below)\n"
+        "           q       .. select genotypes using -i/-e options\n"
+        "           r:FLOAT .. select randomly a proportion of FLOAT genotypes (can be combined with other modes)\n"
         "       and the new genotype can be one of:\n"
-        "           .    .. missing (\".\" or \"./.\", keeps ploidy)\n"
-        "           0    .. reference allele (e.g. 0/0 or 0, keeps ploidy)\n"
-        "           c:GT .. custom genotype (e.g. 0/0, 0, 0/1, m/M, overrides ploidy)\n"
-        "           m    .. minor (the second most common) allele (e.g. 1/1 or 1, keeps ploidy)\n"
-        "           M    .. major allele (e.g. 1/1 or 1, keeps ploidy)\n"
-        "           p    .. phase genotype (0/1 becomes 0|1)\n"
-        "           u    .. unphase genotype and sort by allele (1|0 becomes 0/1)\n"
+        "           .       .. missing (\".\" or \"./.\", keeps ploidy)\n"
+        "           0       .. reference allele (e.g. 0/0 or 0, keeps ploidy)\n"
+        "           c:GT    .. custom genotype (e.g. 0/0, 0, 0/1, m/M, 0/X overrides ploidy)\n"
+        "           m       .. minor (the second most common) allele as determined from INFO/AC or FMT/GT (e.g. 1/1 or 1, keeps ploidy)\n"
+        "           M       .. major allele as determined from INFO/AC or FMT/GT (e.g. 1/1 or 1, keeps ploidy)\n"
+        "           X       .. allele with bigger read depth as determined from FMT/AD\n"
+        "           p       .. phase genotype (0/1 becomes 0|1)\n"
+        "           u       .. unphase genotype and sort by allele (1|0 becomes 0/1)\n"
         "Usage: bcftools +setGT [General Options] -- [Plugin Options]\n"
         "Options:\n"
         "   run \"bcftools plugin\" for a list of common options\n"
         "\n"
         "Plugin options:\n"
-        "   -e, --exclude <expr>        Exclude a genotype if true (requires -t q)\n"
-        "   -i, --include <expr>        include a genotype if true (requires -t q)\n"
-        "   -n, --new-gt <type>         Genotypes to set, see above\n"
-        "   -t, --target-gt <type>      Genotypes to change, see above\n"
+        "   -e, --exclude EXPR        Exclude a genotype if true (requires -t q)\n"
+        "   -i, --include EXPR        include a genotype if true (requires -t q)\n"
+        "   -n, --new-gt TYPE         Genotypes to set, see above\n"
+        "   -s, --seed INT            Random seed to use with -t r [0]\n"
+        "   -t, --target-gt TYPE      Genotypes to change, see above\n"
         "\n"
         "Example:\n"
         "   # set missing genotypes (\"./.\") to phased ref genotypes (\"0|0\")\n"
@@ -165,7 +172,7 @@ void parse_binom_expr(args_t *args, char *str)
     memcpy(args->binom_tag,beg,end-beg);
     int tag_id = bcf_hdr_id2int(args->in_hdr,BCF_DT_ID,args->binom_tag);
     if ( !bcf_hdr_idinfo_exists(args->in_hdr,BCF_HL_FMT,tag_id) ) error("The FORMAT tag \"%s\" is not present in the VCF\n", args->binom_tag);
-    
+
     while ( *end && isspace(*end) ) end++;
     if ( !*end ) _parse_binom_expr_error(str);
 
@@ -194,6 +201,7 @@ int init(int argc, char **argv, bcf_hdr_t *in, bcf_hdr_t *out)
     args->in_hdr  = in;
     args->out_hdr = out;
 
+    char *tmp;
     int c;
     static struct option loptions[] =
     {
@@ -201,12 +209,17 @@ int init(int argc, char **argv, bcf_hdr_t *in, bcf_hdr_t *out)
         {"exclude",required_argument,NULL,'e'},
         {"new-gt",required_argument,NULL,'n'},
         {"target-gt",required_argument,NULL,'t'},
+        {"seed",required_argument,NULL,'s'},
         {NULL,0,NULL,0}
     };
-    while ((c = getopt_long(argc, argv, "?hn:t:i:e:",loptions,NULL)) >= 0)
+    while ((c = getopt_long(argc, argv, "?hn:t:i:e:s:",loptions,NULL)) >= 0)
     {
-        switch (c) 
+        switch (c)
         {
+            case 's':
+                args->rand_seed = strtol(optarg,&tmp,10);
+                if ( *tmp ) error("Could not parse: -s %s\n",optarg);
+                break;
             case 'e':
                 if ( args->filter_str ) error("Error: only one -i or -e expression can be given, and they cannot be combined\n");
                 args->filter_str = optarg; args->filter_logic |= FLT_EXCLUDE; break;
@@ -218,6 +231,7 @@ int init(int argc, char **argv, bcf_hdr_t *in, bcf_hdr_t *out)
                 if ( strchr(optarg,'0') ) args->new_mask |= GT_REF;
                 if ( strchr(optarg,'m') ) args->new_mask |= GT_MINOR;
                 if ( strchr(optarg,'M') ) args->new_mask |= GT_MAJOR;
+                if ( strchr(optarg,'X') ) args->new_mask |= GT_X_VAF;
                 if ( strchr(optarg,'p') ) args->new_mask |= GT_PHASED;
                 if ( strchr(optarg,'u') ) args->new_mask |= GT_UNPHASED;
                 if ( !strncmp(optarg,"c:",2) ) { args->new_mask |= GT_CUSTOM; args->custom.gt_str = optarg; }
@@ -230,6 +244,13 @@ int init(int argc, char **argv, bcf_hdr_t *in, bcf_hdr_t *out)
                 if ( !strcmp(optarg,"a") ) args->tgt_mask |= GT_ALL;
                 if ( !strcmp(optarg,"q") ) args->tgt_mask |= GT_QUERY;
                 if ( !strcmp(optarg,"?") ) args->tgt_mask |= GT_QUERY;        // for backward compatibility
+                if ( !strncmp(optarg,"r:",2) )
+                {
+                    args->rand_frac = strtod(optarg+2,&tmp);
+                    if ( *tmp ) error("Could not parse: -t %s\n", optarg);
+                    if ( args->rand_frac<=0 || args->rand_frac>=1 ) error("Expected value between 0 and 1 with -t\n");
+                    args->tgt_mask |= GT_RAND;
+                }
                 if ( strchr(optarg,'b') ) parse_binom_expr(args, strchr(optarg,'b'));
                 if ( args->tgt_mask==0 ) error("Unknown parameter to --target-gt: %s\n", optarg);
                 break;
@@ -241,6 +262,11 @@ int init(int argc, char **argv, bcf_hdr_t *in, bcf_hdr_t *out)
 
     if ( !args->new_mask ) error("Expected -n option\n");
     if ( !args->tgt_mask ) error("Expected -t option\n");
+    if ( args->tgt_mask & GT_RAND )
+    {
+        if ( args->tgt_mask==GT_RAND ) args->tgt_mask |= GT_ALL;
+        hts_srand48(args->rand_seed);
+    }
 
     if ( args->new_mask & GT_MISSING ) args->new_gt = bcf_gt_missing;
     if ( args->new_mask & GT_REF ) args->new_gt = args->new_mask&GT_PHASED ? bcf_gt_phased(0) : bcf_gt_unphased(0);
@@ -252,6 +278,7 @@ int init(int argc, char **argv, bcf_hdr_t *in, bcf_hdr_t *out)
             int allele;
             if ( *ptr=='m' ) { allele = MINOR_ALLELE; args->new_mask |= GT_MINOR; }
             else if ( *ptr=='M' ) { allele = MAJOR_ALLELE; args->new_mask |= GT_MAJOR; }
+            else if ( *ptr=='X' ) { allele = X_VAF_ALLELE; args->new_mask |= GT_X_VAF; }
             else if ( *ptr=='/' || *ptr=='|' )
             {
                 if ( !args->custom.ploidy ) error("Could not parse the genotype: %s\n",args->custom.gt_str);
@@ -287,10 +314,13 @@ int init(int argc, char **argv, bcf_hdr_t *in, bcf_hdr_t *out)
     int id = bcf_hdr_id2int(args->in_hdr,BCF_DT_ID,"GT");
     if ( !bcf_hdr_idinfo_exists(args->in_hdr,BCF_HL_FMT,id) )
         bcf_hdr_printf(args->out_hdr, "##FORMAT=<ID=GT,Number=1,Type=String,Description=\"Genotype\">");
+    if ( (args->new_mask & GT_X_VAF) && !bcf_hdr_idinfo_exists(args->in_hdr,BCF_HL_FMT,bcf_hdr_id2int(args->in_hdr,BCF_DT_ID,"AD")) )
+        error("Error: the FORMAT/AD annotation does exist, cannot run with --new-gt %s\n",args->custom.gt_str);
 
     return 0;
 }
 
+// sets GT phase for a single sample, ngts is the ploidy
 static inline int phase_gt(int32_t *ptr, int ngts)
 {
     int j, changed = 0;
@@ -304,6 +334,7 @@ static inline int phase_gt(int32_t *ptr, int ngts)
     return changed;
 }
 
+// unphase GT for a single sample, ngts is the ploidy
 static inline int unphase_gt(int32_t *ptr, int ngts)
 {
     int j, changed = 0;
@@ -330,17 +361,21 @@ static inline int unphase_gt(int32_t *ptr, int ngts)
     }
     return changed;
 }
-static inline int set_gt(int32_t *ptr, int ngts, int gt)
+
+// sets GT for a single sample, ngts is the ploidy, allele
+static inline int set_gt(int32_t *ptr, int ngts, int allele)
 {
     int j, changed = 0;
     for (j=0; j<ngts; j++)
     {
         if ( ptr[j]==bcf_int32_vector_end ) break;
-        if ( ptr[j] != gt ) changed++;
-        ptr[j] = gt;
+        if ( ptr[j] != allele ) changed++;
+        ptr[j] = allele;
     }
     return changed;
 }
+
+// sets GT for a single sample using custom.[mMX]_allele; ngts is the ploidy, nals is the number of REF+ALT alleles
 static inline int set_gt_custom(args_t *args, int32_t *ptr, int ngts, int nals)
 {
     int i, changed = 0, new_allele = 0;
@@ -348,29 +383,27 @@ static inline int set_gt_custom(args_t *args, int32_t *ptr, int ngts, int nals)
     {
         if ( args->custom.gt[i]==MINOR_ALLELE ) new_allele = args->custom.m_allele;
         else if ( args->custom.gt[i]==MAJOR_ALLELE ) new_allele = args->custom.M_allele;
+        else if ( args->custom.gt[i]==X_VAF_ALLELE ) new_allele = args->custom.x_vaf_allele==bcf_gt_missing ? nals : bcf_gt_allele(args->custom.x_vaf_allele);
         else new_allele = args->custom.gt[i];
-        if ( new_allele >= nals ) continue;
-        new_allele = args->custom.phased[i] ? bcf_gt_phased(new_allele) : bcf_gt_unphased(new_allele);
+        if ( new_allele >= nals ) // cannot set, the requested index is bigger than there are alleles in ALT
+            new_allele = bcf_gt_missing;
+        else
+            new_allele = args->custom.phased[i] ? bcf_gt_phased(new_allele) : bcf_gt_unphased(new_allele);
         if ( ptr[i]==new_allele ) continue;
         ptr[i] = new_allele;
         changed = 1;
     }
     while ( i<ngts )
+    {
+        if ( !changed && ptr[i]!=bcf_int32_vector_end ) changed = 1;
         ptr[i++] = bcf_int32_vector_end;
+    }
     return changed;
 }
 
-static inline double calc_binom(int na, int nb)
+static inline int random_draw(args_t *args)
 {
-    if ( na + nb == 0 ) return 1;
-
-    /*
-        kfunc.h implements kf_betai, which is the regularized beta function I_x(a,b) = P(X<=a/(a+b))
-    */
-    double prob = na > nb ? 2*kf_betai(na, nb + 1, 0.5) : 2*kf_betai(nb, na + 1, 0.5);
-    if ( prob > 1 ) prob = 1;
-
-    return prob;
+    return hts_drand48() > args->rand_frac ? 1 : 0; // reversed random draw
 }
 
 bcf1_t *process(bcf1_t *rec)
@@ -401,14 +434,6 @@ bcf1_t *process(bcf1_t *rec)
         }
     }
 
-    int nbinom = 0;
-    if ( args->tgt_mask & GT_BINOM )
-    {
-        nbinom = bcf_get_format_int32(args->in_hdr, rec, args->binom_tag, &args->iarr, &args->miarr);
-        if ( nbinom<0 ) nbinom = 0;
-        nbinom /= rec->n_sample;
-    }
-    
     // Calculating allele frequency for each allele and determining major allele
     // only do this if use_major is true
     if ( args->new_mask & GT_MAJOR )
@@ -450,6 +475,35 @@ bcf1_t *process(bcf1_t *rec)
         args->new_gt = args->new_mask & GT_PHASED ?  bcf_gt_phased(imax2) : bcf_gt_unphased(imax2);
         if ( args->new_mask & GT_CUSTOM ) args->custom.m_allele = imax2;
     }
+    if ( args->new_mask & GT_X_VAF )
+    {
+        if ( bcf_get_format_int32(args->in_hdr,rec,"AD",&args->xarr,&args->mxarr)==rec->n_allele*rec->n_sample )
+        {
+            int32_t *src = args->xarr, *dst = args->xarr;
+            for (i=0; i<rec->n_sample; i++)
+            {
+                int jmax = -1;
+                for (j=0; j<rec->n_allele; j++)
+                {
+                    if ( src[j]==bcf_int32_vector_end ) break;
+                    if ( src[j]==bcf_int32_missing  ) continue;
+                    if ( j==0 || src[jmax] < src[j] ) jmax = j;
+                }
+                *dst = jmax==-1 ? bcf_gt_missing : bcf_gt_unphased(jmax);
+                src += rec->n_allele;
+                dst++;
+            }
+        }
+        else return rec;
+    }
+
+    int nbinom = 0;
+    if ( args->tgt_mask & GT_BINOM )
+    {
+        nbinom = bcf_get_format_int32(args->in_hdr, rec, args->binom_tag, &args->iarr, &args->miarr);
+        if ( nbinom<0 ) nbinom = 0;
+        nbinom /= rec->n_sample;
+    }
 
     // replace gts
     if ( nbinom && ngts>=2 )    // only diploid genotypes are considered: higher ploidy ignored further, haploid here
@@ -465,21 +519,27 @@ bcf1_t *process(bcf1_t *rec)
             int32_t *ptr = args->gts + i*ngts;
             if ( bcf_gt_is_missing(ptr[0]) || bcf_gt_is_missing(ptr[1]) || ptr[1]==bcf_int32_vector_end ) continue;
             if ( ptr[0]==ptr[1] ) continue; // a hom
-            int ia = bcf_gt_allele(ptr[0]); 
-            int ib = bcf_gt_allele(ptr[1]); 
-            if ( ia>=nbinom || ib>=nbinom ) 
+            int ia = bcf_gt_allele(ptr[0]);
+            int ib = bcf_gt_allele(ptr[1]);
+            if ( ia>=nbinom || ib>=nbinom )
                 error("The sample %s has incorrect number of %s fields at %s:%"PRId64"\n",
                         args->in_hdr->samples[i],args->binom_tag,bcf_seqname(args->in_hdr,rec),(int64_t) rec->pos+1);
 
-            double prob = calc_binom(args->iarr[i*nbinom+ia],args->iarr[i*nbinom+ib]);
+            double prob = calc_binom_two_sided(args->iarr[i*nbinom+ia],args->iarr[i*nbinom+ib],0.5);
             if ( !args->binom_cmp(prob,args->binom_val) ) continue;
+            if ( args->tgt_mask&GT_RAND && random_draw(args) ) continue;
 
             if ( args->new_mask&GT_UNPHASED )
                 changed += unphase_gt(ptr, ngts);
             else if ( args->new_mask==GT_PHASED )
                 changed += phase_gt(ptr, ngts);
             else if ( args->new_mask & GT_CUSTOM )
+            {
+                if ( args->new_mask & GT_X_VAF ) args->custom.x_vaf_allele = args->xarr[i];
                 changed += set_gt_custom(args, ptr, ngts, rec->n_allele);
+            }
+            else if ( args->new_mask & GT_X_VAF )
+                changed += set_gt(ptr, ngts, args->xarr[i]);
             else
                 changed += set_gt(ptr, ngts, args->new_gt);
         }
@@ -509,12 +569,18 @@ bcf1_t *process(bcf1_t *rec)
         for (i=0; i<rec->n_sample; i++)
         {
             if ( args->smpl_pass && !args->smpl_pass[i] ) continue;
+            if ( args->tgt_mask&GT_RAND && random_draw(args) ) continue;
             if ( args->new_mask&GT_UNPHASED )
                 changed += unphase_gt(args->gts + i*ngts, ngts);
             else if ( args->new_mask==GT_PHASED )
                 changed += phase_gt(args->gts + i*ngts, ngts);
             else if ( args->new_mask & GT_CUSTOM )
+            {
+                if ( args->new_mask & GT_X_VAF ) args->custom.x_vaf_allele = args->xarr[i];
                 changed += set_gt_custom(args, args->gts + i*ngts, ngts, rec->n_allele);
+            }
+            else if ( args->new_mask & GT_X_VAF )
+                changed += set_gt(args->gts + i*ngts, ngts, args->xarr[i]);
             else
                 changed += set_gt(args->gts + i*ngts, ngts, args->new_gt);
         }
@@ -529,7 +595,7 @@ bcf1_t *process(bcf1_t *rec)
             {
                 if ( ptr[j]==bcf_int32_vector_end ) break;
                 ploidy++;
-                if ( ptr[j]==bcf_gt_missing ) nmiss++;
+                if ( bcf_gt_is_missing(ptr[j]) ) nmiss++;
             }
 
             int do_set = 0;
@@ -538,13 +604,19 @@ bcf1_t *process(bcf1_t *rec)
             else if ( args->tgt_mask&GT_MISSING && ploidy==nmiss ) do_set = 1;
 
             if ( !do_set ) continue;
+            if ( args->tgt_mask&GT_RAND && random_draw(args) ) continue;
 
             if ( args->new_mask&GT_UNPHASED )
                 changed += unphase_gt(ptr, ngts);
             else if ( args->new_mask==GT_PHASED )
                 changed += phase_gt(ptr, ngts);
             else if ( args->new_mask & GT_CUSTOM )
+            {
+                if ( args->new_mask & GT_X_VAF ) args->custom.x_vaf_allele = args->xarr[i];
                 changed += set_gt_custom(args, args->gts + i*ngts, ngts, rec->n_allele);
+            }
+            else if ( args->new_mask & GT_X_VAF )
+                changed += set_gt(ptr, ngts, args->xarr[i]);
             else
                 changed += set_gt(ptr, ngts, args->new_gt);
         }
@@ -567,6 +639,7 @@ void destroy(void)
     if ( args->filter ) filter_destroy(args->filter);
     free(args->arr);
     free(args->iarr);
+    free(args->xarr);
     free(args->gts);
     free(args);
 }
